// Implementation of the main sq3 classes.
//
//
//////////////////////////////////////////////////////////////////////

#include "SQ3.h"

namespace sq3 {

	// SQ3_ASSERT_DB(FuncName,RetVal) is used only be the Database class. It asserts
	// that this->m_db is non-0. If that fails then RetVal is returned. For void functions,
	// pass SQ3_VOID as the return type (a lame kludge to accommodate an MSVC bug). In
	// real compilers we can use 'return void', but MSVC won't allow that here, nor will
	// it allow the equally-legal 'return (void)0', and thus the SQ3_VOID kludge.
#define SQ3_ASSERT_DB(FuncName,RetVal)					\
	if( ! this->m_db ) {						\
		CString assertion; assertion.Format( _T("Database::%s: db not opened."), CString(# FuncName) ); \
		AfxMessageBox(assertion);				\
		return RetVal;						\
	}
#define SQ3_VOID // placeholder to allow us to return void from SQ3_ASSERT_DB

	/**
	   Helper routine to get a non-const (char *) from CStrings.

	   input = the input string.  targetbuffer must be allocated
	   (and deallocated) by the caller and must be at least
	   (maxlength+1) bytes long (+1 for a trailing null).

	   input is copied to targetbuffer and padded with a null
	   character.

	   Returns targetbuffer.

	   There is absolutely no guaranty that this function will work
	   for input strings containing non-latin1 characters.

	   Tip: use std::vector<char> as a (char *) buffer to avoid
	   having to allocate/deallocate the string yourself.
	*/
	static char *CString2CharString(CString const &input, char *targetbuffer, int maxlength)
	{
		int i;
		const int inlen = input.GetLength();
		for(i=0;i<inlen;i++)
		{
			if(i>=(maxlength-1)) break; // not bigger than maxlength - 1
			targetbuffer[i] = (char) input.GetAt(i);
		}

		targetbuffer[i] = '\0';
		targetbuffer[maxlength-1] = '\0'; // just to be sure

		return targetbuffer;
	}


	
    
	CharStringConverter::CharStringConverter( CString const & s )
		: m_len( (unsigned int) s.GetLength()),
		  m_buf(0)
	{
		if( this->m_len > 0 )
		{
			this->m_buf = new char[this->m_len+1];
			CString2CharString( s, this->m_buf, this->m_len+1 );
		}
	}


	CharStringConverter::~CharStringConverter()
	{
		if( this->m_buf )
		{
			delete [] this->m_buf;
		}
	}

	char * CharStringConverter::CharString()
	{
		return this->m_buf;
	}

	char const * CharStringConverter::CharString() const
	{
		return this->m_buf;
	}

	CharStringConverter::operator char *()
	{
		return this->m_buf;
	}

	CharStringConverter::operator char const *() const
	{
		return this->m_buf;
	}


	unsigned int CharStringConverter::GetLength() const
	{
		return this->m_len;
	}


	Database::Database( CString const & n)
		: m_db(0),
		  m_name(n)
	{
		this->Open();
	}
    
	Database::~Database()
	{
		this->Close();
	}

	int Database::Close()
	{
		if( ! this->m_db ) return 0;
		int rc = sqlite3_close( this->m_db );
		this->m_db = 0; // should we only do this is close succeeds?
		return rc;
	}

	int Database::Open( CString const & name )
	{
		this->Close();
		this->m_name = name;
		return sqlite3_open16( name, &this->m_db );
	}

	int Database::Open()
	{
		return this->Open( this->m_name );
	}
    
	bool Database::IsOpen() const
	{
		return 0 != this->m_db;
	}

	CString Database::GetName() const
	{
		return this->m_name;
	}
	sqlite3 * Database::GetHandle() const
	{
		return this->m_db;
	}

	CString Database::GetErrorMessage() const
	{
		SQ3_ASSERT_DB(GetErrorMessage,CString());
		char const * s = sqlite3_errmsg( this->m_db );
		return s ? CString(s) : CString();
	}

	int64_t Database::GetLastInsertID() const
	{
		SQ3_ASSERT_DB(GetLastInsertID,SQLITE_ERROR);
		return sqlite3_last_insert_rowid(this->m_db);
	}

	void Database::SetBusyTimeout( int millis )
	{
		SQ3_ASSERT_DB(SetBusyTimeout,SQ3_VOID);
		sqlite3_busy_timeout(this->m_db, millis);
	}

	int Database::Execute( char const * sql, sqlite3_callback callback, void * data )
	{
		SQ3_ASSERT_DB(Execute,SQLITE_ERROR);
		return sqlite3_exec( this->m_db, sql, callback, data, 0 );
	}

	int Database::Execute( CString const & sql, sqlite3_callback callback, void * data )
	{
		SQ3_ASSERT_DB(Execute,SQLITE_ERROR);
		CharStringConverter conv(sql);
		return this->Execute( conv.CharString(), callback, data );
	}

    
	int Database::ExecuteInt( CString const & sql, int & val )
	{
		SQ3_ASSERT_DB(ExecuteInt,SQLITE_ERROR);
		return Statement( *this, sql ).ExecuteInt( val );
	}

	int Database::ExecuteInt64( CString const & sql, int64_t & val )
	{
		SQ3_ASSERT_DB(ExecuteInt64,SQLITE_ERROR);
		return Statement( *this, sql ).ExecuteInt64( val );
	}

	int Database::ExecuteDouble( CString const & sql, double & val )
	{
		SQ3_ASSERT_DB(ExecuteDouble,SQLITE_ERROR);
		return Statement( *this, sql ).ExecuteDouble( val );
	}


	int Database::ExecuteString( CString const & sql, CString & val )
	{
		SQ3_ASSERT_DB(ExecuteString,SQLITE_ERROR);
		return Statement( *this, sql ).ExecuteString( val );
	}


	void const * Database::ExecuteBlob( char const * sql, int & size )
	{
		SQ3_ASSERT_DB(ExecuteBlob,(size=-1,0));
		return Statement( *this, sql ).ExecuteBlob( size );
	}


	Transaction::Transaction(Database &con, bool start)
		: m_con(con),m_intrans(false)
	{
		if(start) this->Begin();
	}

	Transaction::~Transaction()
	{
		if(this->m_intrans)
		{
			__try
			{
				this->Rollback();
			}
			__except(EXCEPTION_EXECUTE_HANDLER)
			{
				return;
			}
		}
	}

	int Transaction::Begin()
	{
		if( this->m_intrans )
		{
			return SQLITE_ERROR;
		}
		int rc = this->m_con.Execute("begin;");
		this->m_intrans = (0 == rc);
		return rc;
	}

	int Transaction::Commit()
	{
		if( ! this->m_intrans )
		{
			return SQLITE_ERROR;
		}
		int rc = this->m_con.Execute("commit;");
		this->m_intrans = (0 == rc);
		return rc;
	}

	int Transaction::Rollback()
	{
		int rc = this->m_con.Execute("rollback;");
		this->m_intrans = false;
		return rc;
	}
   
	bool Transaction::IsActive() const
	{
		return this->m_intrans;
	}

	Reader::Reader() : m_cmd(0)
	{

	}

	Reader::Reader( Statement * s ) : m_cmd(s)
	{
		++this->m_cmd->m_refs;
	}

	Reader & Reader::operator=( Reader const & rhs )
	{
		if( &rhs != this )
		{
			//this->Close();
			this->m_cmd = rhs.m_cmd;
			if( this->m_cmd )
			{
				++this->m_cmd->m_refs;
			}
		}
		return *this;
	}
	Reader::Reader( Reader const & rhs )
		: m_cmd(rhs.m_cmd)
	{
		if( this->m_cmd )
		{
			++this->m_cmd->m_refs;
		}
	}


	Reader::~Reader()
	{
		this->Close();
	}

	int Reader::GetColCount()
	{
		return this->m_cmd
			? this->m_cmd->GetColCount()
			: -1;
	}
	int Reader::Step()
	{
		if(!this->m_cmd)
		{
			return SQLITE_ERROR;
		}
		return sqlite3_step(this->m_cmd->m_st);
	}

	int Reader::Reset()
	{
		return this->m_cmd
			? this->m_cmd->Reset()
			: SQLITE_ERROR;
	}

	int Reader::Close()
	{
		if(! this->m_cmd)
		{
			return 0;
		}
		int rc = SQLITE_OK;
		if(--this->m_cmd->m_refs==0)
		{
			rc = this->m_cmd->Reset();
		}
		this->m_cmd = 0;
		return rc;
	}

	bool Reader::IsValidIndex( int index )
	{
		return this->m_cmd &&
			( index < this->m_cmd->m_colcount);
	}

	CString Reader::GetColName( int index )
	{
		return this->IsValidIndex( index )
			? sqlite3_column_name( this->m_cmd->m_st, index )
			: CString();
	}

	bool Reader::GetInt( int index, int & val )
	{
		if( ! this->IsValidIndex( index ) )
		{
			return false;
		}
		val = sqlite3_column_int( this->m_cmd->m_st, index );
		return true;
	}

	bool Reader::GetInt64( int index, int64_t & val )
	{
		if( ! this->IsValidIndex( index ) )
		{
			return false;
		}
		val = sqlite3_column_int64( this->m_cmd->m_st, index );
		return true;
	}

	bool Reader::GetDouble( int index, double & val )
	{
		if( ! this->IsValidIndex( index ) )
		{
			return false;
		}
		val = sqlite3_column_double( this->m_cmd->m_st, index );
		return true;
	}

	bool Reader::GetString( int index, CString & val )
	{
		if( ! this->IsValidIndex( index ) )
		{
			return false;
		}
		val = (LPCTSTR) sqlite3_column_text16( this->m_cmd->m_st, index );
		return true;
	}

	void const * Reader::GetBlob( int index, int & size )
	{
		if( ! this->IsValidIndex( index ) )
		{
			size = -1;
			return 0;
		}
		size = sqlite3_column_bytes( this->m_cmd->m_st, index );
		return sqlite3_column_blob( this->m_cmd->m_st, index );
	}

	Statement::Statement( Database &con, CString const & sql )
		: m_db(&con), m_st(0), m_refs(0), m_colcount(0)
	{
		void const * tail = 0;
		if( SQLITE_OK == sqlite3_prepare16( this->m_db->GetHandle(), sql, -1, &this->m_st, &tail ) )
		{
			this->m_colcount = sqlite3_column_count( this->m_st );
		}
	}
	Statement::~Statement()
	{
		this->Finalize();
	}

	int Statement::Finalize()
	{
		int rc = SQLITE_OK;
		if( this->m_st )
		{
			rc = sqlite3_finalize( this->m_st );
			this->m_st = 0;
		}
		return rc;
	}

	int Statement::Reset()
	{
		return this->m_st
			? sqlite3_reset( this->m_st )
			: SQLITE_ERROR;
	}

	bool Statement::IsValid() const
	{
		return 0 != this->m_st;
	}

	int Statement::Bind( int index )
	{
		return this->IsValid()
			? sqlite3_bind_null( this->m_st, index )
			: SQLITE_ERROR;
	}
	int Statement::Bind( int index, int data )
	{
		return this->IsValid()
			? sqlite3_bind_int( this->m_st, index, data )
			: SQLITE_ERROR;
	}
    
	int Statement::Bind( int index, int64_t data )
	{
		return this->IsValid()
			? sqlite3_bind_int64( this->m_st, index, data )
			: SQLITE_ERROR;
	}
    
	int Statement::Bind( int index, double data )
	{
		return this->IsValid()
			? sqlite3_bind_double( this->m_st, index, data )
			: SQLITE_ERROR;
	}
	int Statement::Bind( int index, CString const & data )
	{
		return this->IsValid()
			? sqlite3_bind_text16( this->m_st, index, data, -1, SQLITE_TRANSIENT )
			: SQLITE_ERROR;
	}
    
	int Statement::Bind( int index, char const * data, int datalen )
	{
		return this->IsValid()
			? sqlite3_bind_text( this->m_st, index, data, datalen, SQLITE_TRANSIENT )
			: SQLITE_ERROR;
	}
    
	int Statement::Bind( int index, void const * data, int datalen )
	{
		return this->IsValid()
			? sqlite3_bind_blob( this->m_st, index, data, datalen, SQLITE_TRANSIENT )
			: SQLITE_ERROR;
	}
    
	Reader Statement::ExecuteReader()
	{
		return Reader(this);
	}
	int Statement::ExecuteNonQuery()
	{
		return this->ExecuteReader().Step();
	}
	int Statement::ExecuteInt( int & val )
	{
		Reader r(this->ExecuteReader());
		if(SQLITE_ROW != r.Step())
		{
			return SQLITE_ERROR;
		}
		return r.GetInt( 0, val )
			? SQLITE_OK
			: SQLITE_ERROR;
	}

	int Statement::ExecuteInt64( int64_t & val)
	{
		Reader r(this->ExecuteReader());
		if(SQLITE_ROW != r.Step())
		{
			return SQLITE_ERROR;
		}
		return r.GetInt64( 0, val )
			? SQLITE_OK
			: SQLITE_ERROR;
	}
	int Statement::ExecuteDouble( double & val )
	{
		Reader r(this->ExecuteReader());
		if(SQLITE_ROW != r.Step())
		{
			return SQLITE_ERROR;
		}
		return r.GetDouble( 0, val )
			? SQLITE_OK
			: SQLITE_ERROR;
	}
	int Statement::ExecuteString( CString & val)
	{
		Reader r(this->ExecuteReader());
		if(SQLITE_ROW != r.Step())
		{
			return SQLITE_ERROR;
		}
		return r.GetString( 0, val )
			? SQLITE_OK
			: SQLITE_ERROR;
	}
	void const * Statement::ExecuteBlob( int & size )
	{
		Reader r(this->ExecuteReader());
		if(SQLITE_ROW != r.Step())
		{
			size = -1;
			return 0;
		}
		return r.GetBlob( 0, size );
	}

	int Statement::GetColCount() const
	{
		return this->m_colcount;
	}
} // namespace

#undef SQ3_ASSERT_DB
#undef SQ3_VOID
